import { Bytes, Timestamp as Timestamp$1 } from 'firebase/firestore';
import { Timestamp as Timestamp$2 } from 'firebase-admin/firestore';

/**
 * Type alias for representing binary data in Firestore. Can be Bytes from
 * the Firebase SDK or Uint8Array (/ )Buffer) when running in Node.js
 */
type Binary = Bytes | Uint8Array;
/**
 * Type alias for a Firestore Timestamp value. Can represent either
 * a Timestamp from the client SDK or admin SDK. Allows writing
 * code that works with both.
 */
type Timestamp = Timestamp$1 | Timestamp$2;
/** Type alias for Firestore sentinel values */
type FieldValue = any;
/**
 * Exports the Adapter interface which defines methods for converting
 * between Firestore data types like Binary and Timestamp and other formats.
 *
 * This allows writing code that is agnostic to whether it is running
 * in a Node.js or browser context.
 */
interface Adapter {
    /**
     * Converts a base64 encoded string to a Binary value.
     */
    fromBase64String(value: string): Binary;
    /**
     * Converts a Binary value to a base64 encoded string.
     */
    toBase64String(value: Binary): string;
    /**
     * Converts a Uint8Array to a Binary value.
     */
    fromUint8Array(value: Uint8Array): Binary;
    /**
     * Converts a Binary value to a Uint8Array.
     */
    toUInt8Array(value: Binary): Uint8Array;
    /**
     * Converts a hexadecimal string to a Binary value.
     */
    fromHexString(value: string): Binary;
    /**
     * Converts a Binary value to a hexadecimal string representation.
     */
    toHexString(value: Binary): string;
    /**
     * Converts a string value to a Binary data type.
     */
    fromString(value: string): Binary;
    /**
     * Converts a Binary value to a string representation.
     */
    toString(value: Binary): string;
    /**
     * Converts a Date value to a Timestamp.
     */
    fromDate(value: Date): Timestamp;
    /**
     * Converts a Timestamp value to a JavaScript Date object.
     */
    toDate(value: Timestamp): Date;
    /**
     * Checks if the given value is a Binary data type.
     */
    isBinary(value: any): boolean;
    /**
     * Checks if the given value is of the Firestore Timestamp type.
     */
    isTimestamp(value: any): boolean;
    /**
     * Returns a FieldValue sentinel to remove the given elements from an array field in a Firestore document.
     */
    arrayRemove(...elements: any[]): FieldValue;
    /**
     * Returns a FieldValue sentinel to add the given elements to an array field in a Firestore document.
     */
    arrayUnion(...elements: any[]): FieldValue;
    /**
     * Returns a FieldValue sentinel to delete the given field from the document on the server.
     */
    delete(): FieldValue;
    /**
     * Returns a FieldValue sentinel to increment the given field's value by the amount specified.
     */
    increment(n: number): FieldValue;
    /**
     * Returns a FieldValue sentinel to set the field to the server's current timestamp.
     */
    serverTimestamp(): FieldValue;
}

/**
 * The DocumentData interface defines the structure of a Firestore document.
 * It contains string keys mapping to values of any type.
 */
interface DocumentData {
    [field: string]: any;
}
interface CollectionReference<T, TDB extends DocumentData = DocumentData> {
    get id(): string;
    get path(): string;
    get parent(): DocumentReference<DocumentData, DocumentData> | null;
}
interface DocumentReference<T, TDB extends DocumentData = DocumentData> {
    get id(): string;
    get path(): string;
    get parent(): CollectionReference<T, TDB>;
}
/**
 * Interface for a Firestore document snapshot, containing the document
 * ID and data. Allows specifying generic types for the data shape T
 * and internal database format TDB.
 */
interface QueryDocumentSnapshot<T, TDB extends DocumentData = DocumentData> {
    get id(): string;
    get ref(): DocumentReference<T, TDB>;
    data(): TDB;
}
/** Type alias for primitive JavaScript types that can be stored in Firestore. */
type Primitive = string | number | boolean | undefined | null;
/**
 * Type alias for wrapping a type T in the Firestore `FieldValue` type.
 * This allows specifying a type to be converted to a Firestore field
 * value like Timestamp or Bytes.
 */
type WithFieldValue<T> = T | (T extends Primitive ? T : T extends {} ? {
    [K in keyof T]: WithFieldValue<T[K]> | FieldValue;
} : never);
/**
 * Interface for a constructor function that creates a FirestoreDataConverter.
 */
interface FirestoreDataConverterConstructor<T, TDB extends DocumentData = DocumentData> {
    new (adapter: Adapter): FirestoreDataConverter<T, TDB>;
}
/**
 * Interface for a FirestoreDataConverter.
 */
interface FirestoreDataConverter<T, TDB extends DocumentData = DocumentData> {
    toFirestore(modelObject: WithFieldValue<T>): WithFieldValue<TDB>;
    fromFirestore(snapshot: QueryDocumentSnapshot<DocumentData, DocumentData>): T;
}
/**
 * Options for configuring behavior of a DefaultConverter.
 */
interface DefaultConverterOptions {
    handle_id?: boolean;
    transform?: (id: string) => string;
}
/**
 * Defines an abstract base class for converters that handle
 * default behavior like transforming IDs.
 */
declare abstract class DefaultConverterBase<T extends DocumentData> implements FirestoreDataConverter<T, DocumentData> {
    private readonly adapter;
    private readonly handle_id;
    private readonly transform;
    constructor(adapter: Adapter, options?: DefaultConverterOptions);
    toFirestore(model: WithFieldValue<T>): WithFieldValue<DocumentData>;
    fromFirestore(snapshot: QueryDocumentSnapshot<T, DocumentData>): T;
}

export { type Adapter as A, type Binary as B, DefaultConverterBase as D, type FirestoreDataConverterConstructor as F, type QueryDocumentSnapshot as Q, type Timestamp as T, type WithFieldValue as W, type DefaultConverterOptions as a, type DocumentData as b, type FirestoreDataConverter as c };
